<?php

declare(strict_types=1);

namespace Drupal\search_api_typesense_ai\Form;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\search_api_typesense\Entity\TypesenseSchema;
use Drupal\search_api_typesense_ai\EmbeddingModelPluginManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form alter for the typesense_schema config entity.
 */
class SchemaFormAlter implements ContainerInjectionInterface {

  /**
   * SchemaFormAlter constructor.
   *
   * @param \Drupal\search_api_typesense_ai\EmbeddingModelPluginManager $embeddingModelManager
   *   Embedding model plugin manager.
   */
  final public function __construct(
    private readonly EmbeddingModelPluginManager $embeddingModelManager,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): static {
    return new static(
      $container->get('plugin.manager.search_api_typesense_ai.embedding_model'),
    );
  }

  /**
   * Implements hook_form_alter().
   */
  public function formAlter(
    array &$form,
    FormStateInterface $form_state,
    string $form_id,
  ): void {
    if ($form_id == 'typesense_schema_form') {
      /** @var \Drupal\search_api_typesense\Form\SchemaForm $form_object */
      $form_object = $form_state->getFormObject();
      /** @var \Drupal\search_api_typesense\Entity\TypesenseSchema $schema */
      $schema = $form_object->getEntity();

      $form['ai_header'] = [
        '#markup' => '<h2>' . \t('AI features') . '</h2>',
      ];

      $form['enable_embedding'] = [
        '#type' => 'checkbox',
        '#title' => 'Enable embedding',
        '#default_value' => $schema
          ->getThirdPartySetting(
            module: 'search_api_typesense_ai',
            key   : 'enable_embedding',
        ) ?? FALSE,
      ];

      $form['embedding_fields'] = [
        '#type' => 'checkboxes',
        '#title' => \t('Embedding fields'),
        '#description' => \t('List of fields used to create embeddings. Only string and string[] fields are allowed.'),
        '#options' => $schema->getStringFields(),
        '#default_value' => $schema
          ->getThirdPartySetting(
            module: 'search_api_typesense_ai',
            key   : 'embedding_fields',
        ) ?? [],
        '#states' => [
          'invisible' => [
            ':input[name="enable_embedding"]' => [
              'checked' => FALSE,
            ],
          ],
        ],
      ];

      $embedding_model = $schema
        ->getThirdPartySetting(
          module: 'search_api_typesense_ai',
          key   : 'embedding_model',
        ) ?? NULL;

      $form['embedding_model'] = [
        '#type' => 'select',
        '#title' => \t('LLM model'),
        '#description' => \t('The LLM model to be used for embeddings.'),
        '#options' => $this->getAvailableEmbeddingModels(),
        '#default_value' => $embedding_model,
        '#ajax' => [
          'callback' => [SchemaFormAlter::class, 'buildAjaxEmbeddingModelConfigForm'],
          'wrapper' => 'embedding-model-config-form',
          'effect' => 'fade',
        ],
        '#states' => [
          'invisible' => [
            ':input[name="enable_embedding"]' => [
              'checked' => FALSE,
            ],
          ],
        ],
      ];

      $form['embedding_model_config'] = [];
      if ($embedding_model != NULL) {
        /** @var \Drupal\search_api_typesense_ai\EmbeddingModelInterface $embedding_model_instance */
        $embedding_model_instance = $this->embeddingModelManager->createInstance($embedding_model);

        $backend_form_state = SubformState::createForSubform($form['embedding_model_config'], $form, $form_state);

        $form['embedding_model_config'] = $embedding_model_instance
          ->buildConfigurationForm($form['embedding_model_config'], $backend_form_state, $schema);
      }
      $form['embedding_model_config'] += [
        '#type' => 'container',
      ];
      $form['embedding_model_config']['#attributes']['id'] = 'embedding-model-config-form';
      $form['embedding_model_config']['#tree'] = TRUE;
      $form['embedding_model_config']['#states'] = [
        'invisible' => [
          ':input[name="enable_embedding"]' => [
            'checked' => FALSE,
          ],
        ],
      ];

      $form['#entity_builders'][] = [
        SchemaFormAlter::class,
        'entityFormBuilder',
      ];
    }
  }

  /**
   * Handles switching the selected backend plugin.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The part of the form to return as AJAX.
   */
  public static function buildAjaxEmbeddingModelConfigForm(
    array $form,
    FormStateInterface $form_state,
  ): array {
    return $form['embedding_model_config'];
  }

  /**
   * Entity builder for the typesense_schema config entity.
   */
  public static function entityFormBuilder(
    string $entity_type,
    TypesenseSchema $schema,
    array &$form,
    FormStateInterface $form_state,
  ): void {
    if ($form_state->getValue('enable_embedding') === 1) {
      $schema->setThirdPartySetting(
        module: 'search_api_typesense_ai',
        key   : 'enable_embedding',
        value : $form_state->getValue('enable_embedding'),
      );

      $schema->setThirdPartySetting(
        module: 'search_api_typesense_ai',
        key   : 'embedding_fields',
        value : $form_state->getValue('embedding_fields'),
      );

      $schema->setThirdPartySetting(
        module: 'search_api_typesense_ai',
        key   : 'embedding_model',
        value : $form_state->getValue('embedding_model'),
      );

      $schema->setThirdPartySetting(
        module: 'search_api_typesense_ai',
        key   : 'embedding_model_config',
        value : $form_state->getValue('embedding_model_config'),
      );

      return;
    }

    $schema->unsetThirdPartySetting(
      module: 'search_api_typesense_ai',
      key   : 'enable_embedding',
    );
  }

  /**
   * Return available embedding models.
   *
   * @return string[]
   *   Available embedding models.
   */
  public function getAvailableEmbeddingModels(): array {
    $models = [];
    $definitions = $this->embeddingModelManager->getDefinitions();
    foreach ($definitions as $id => $definition) {
      $models[$id] = $definition['label'];
    }

    return $models;
  }

}
