<?php

declare(strict_types=1);

namespace Drupal\search_api_typesense\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\search_api\IndexInterface;
use Drupal\search_api_typesense\Plugin\search_api\backend\SearchApiTypesenseBackend;

/**
 * Controller for Typesense index operations.
 */
class TypesenseIndexController extends ControllerBase {

  /**
   * Try the search.
   *
   * @param \Drupal\search_api\IndexInterface $search_api_index
   *   The server.
   *
   * @return array
   *   The render array.
   *
   * @throws \Drupal\search_api\SearchApiException
   */
  public function search(IndexInterface $search_api_index): array {
    $backend = $search_api_index->getServerInstance()->getBackend();
    if (!$backend instanceof SearchApiTypesenseBackend) {
      throw new \InvalidArgumentException('The server must use the Typesense backend.');
    }

    if (!$backend->isAvailable()) {
      $this->messenger()->addError(
        $this->t('The Typesense server is not available.'),
      );

      return [];
    }

    $configuration = $backend->getConfiguration();
    $typesense_client = $backend->getTypesenseClient();
    $all_fields = $typesense_client->getFields($search_api_index->id());
    $query_by_fields = $typesense_client->getFieldsForQueryBy($search_api_index->id());
    $facet_number_fields = $typesense_client->getFieldsForFacetNumber($search_api_index->id());
    $facet_string_fields = $typesense_client->getFieldsForFacetString($search_api_index->id());
    $schema = $backend->getSchema($search_api_index->id());
    $query_by_weights = $schema['fields'] ? $typesense_client->getQueryByWeight($schema['fields']) : '';
    $sort_by_fields = $typesense_client->getFieldsForSortBy($schema['fields']);

    $build = [];
    $build['content'] = [
      '#theme' => 'search_api_typesense_admin_search',
      '#facet_number_fields' => $facet_number_fields,
      '#facet_string_fields' => $facet_string_fields,
      '#attached' => [
        'library' => [
          'search_api_typesense/search',
        ],
        'drupalSettings' => [
          'search_api_typesense' => [
            'api_key' => $configuration['admin_api_key'],
            'host' => $configuration['nodes'][0]['host'],
            'port' => $configuration['nodes'][0]['port'],
            'protocol' => $configuration['nodes'][0]['protocol'],
            'index' => $search_api_index->id(),
            'all_fields' => $all_fields,
            'query_by_fields' => \implode(',', $query_by_fields),
            'query_by_weights' => \implode(',', $query_by_weights),
            'sort_by_fields' => \implode(',', $sort_by_fields),
            'facet_number_fields' => $facet_number_fields,
            'facet_string_fields' => $facet_string_fields,
          ],
        ],
      ],
    ];

    return $build;
  }

}
